# -*- coding: utf-8 -*-
#
# Copyright (C) 2011 UNINETT AS
#
# This file is part of Network Administration Visualized (NAV).
#
# NAV is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License version 2 as published by the Free
# Software Foundation.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
# details.  You should have received a copy of the GNU General Public License
# along with NAV. If not, see <http://www.gnu.org/licenses/>.
#

"""Functions for rendering seeddb list views"""
from collections import defaultdict

from django.core.urlresolvers import reverse
from django.shortcuts import render_to_response
from django.template import RequestContext
from django.db.models.fields import FieldDoesNotExist

from nav.django.utils import get_verbose_name


def render_list(request, queryset, value_list, edit_url=None,
                edit_url_attr='pk', filter_form=None,
                template='seeddb/list.html', extra_context=None,
                censor_list=None, add_descriptions=False):
    """Renders a Seed DB list.

    Parameters:
     - request: HttpRequest object from Django
     - queryset: A queryset containing all objects that should appear in the
                 list.
     - value_list: Tuple with field names used in Django's `value_list`
                   queryset method.
     - edit_url: Named URL to the edit page.
     - edit_url_attr: Attribute used to make the URL to the edit page.
     - filter_form: Form used to filter the queryset.
     - template: Path to the template used.
     - extra_context: A dictionary containing all additional context that
                      should be used in the template.
     - censor_list: list of value headers to censor
    """

    if not extra_context:
        extra_context = {}

    queryset = _filter_query(filter_form, queryset)

    # Get values specified in value_list from the queryset.
    # Also make sure that the primary key and the edit_url_attr appears.
    if not edit_url:
        value_queryset = queryset.values('pk', *value_list)
    else:
        value_queryset = queryset.values('pk', edit_url_attr, *value_list)

    if not edit_url:
        objects, datakeys = _process_objects(value_queryset, value_list)
    else:
        objects, datakeys = _process_objects(value_queryset, value_list,
                                             edit_url, edit_url_attr)

    labels = _label(queryset.model, value_list, datakeys)
    if add_descriptions:
        _add_descriptions(objects, queryset)

    context = {
        'object_list': objects,
        'labels': labels,
        'filter_form': filter_form,
        'sub_active': {'list': True},
        'censor_list': censor_list
    }

    # Update extra_context with context.
    # Doing it this way makes sure that the context dictionary overrides any
    # user-supplied data, not the other way around.
    extra_context.update(context)
    return render_to_response(template, extra_context, RequestContext(request))


def _filter_query(filter_form, queryset):
    """Apply filter_form to queryset.
    """
    if filter_form and filter_form.is_valid():
        filter_data = filter_form.cleaned_data.items()
        filter_tuples = [(key, value) for key, value in filter_data if value]
        query_filter = dict(filter_tuples)
        queryset = queryset.filter(**query_filter)
    return queryset


def _process_objects(query_set_values, value_list, edit_url=None,
                     edit_url_attr=None):
    """Packs values into a format the template understands.

    A list contains each row.
    Each row is a dictionary with following keys:
     - pk: The primary key.
     - url: URL to the edit page.
     - values_list: A list with the values that should be displayed in the
                    table.
    """
    # pick up which values are dictionaries and make note of their existing keys
    datakeys = defaultdict(set)
    for obj in query_set_values:
        for attr in value_list:
            value = obj[attr]
            if isinstance(value, dict):
                datakeys[attr].update(value)
    datakeys = dict([(k, list(sorted(v))) for k, v in datakeys.iteritems()])

    def _getvalues(obj):
        for attr in value_list:
            value = obj[attr]
            if attr in datakeys:
                for key in datakeys[attr]:
                    yield value.get(key, None)
            else:
                yield value

    objects = []
    for obj in query_set_values:
        row = {
            'pk': obj['pk'],
            'values_list': list(_getvalues(obj))
        }
        if edit_url:
            row['url'] = reverse(edit_url, args=(obj[edit_url_attr],))
        objects.append(row)
    return objects, datakeys


def _label(model, value_list, datakeys=None):
    """Make labels for the table head.
    Returns a list of tuples. Each tuple contains the verbose label and a key
    that can be used for sort parameters in the URL.
    """
    attrs = []
    labels = []
    for value in value_list:
        if value in datakeys:
            labels.extend(datakeys[value])
            attrs.extend(datakeys[value])
        else:
            attrs.append(value)
            try:
                labels.append(get_verbose_name(model, value))
            except FieldDoesNotExist:
                labels.append(value)
    return zip(labels, attrs)


def _add_descriptions(objects, queryset):
    """Adds a description key to all objects"""
    for obj in objects:
        model = queryset.get(pk=obj['pk'])
        obj['description'] = getattr(model, 'description', '')


